from django.core.management.base import BaseCommand

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.core import get_bundles, set_bundle_versions
from django_bundles.processors import processor_pipeline, processor_library

import os
import shutil
from hashlib import md5
from tempfile import NamedTemporaryFile

def make_bundle(bundle):
    """
    Does all of the processing required to create a bundle and write it to disk, returning its hash version
    """
    hash_version = None

    with NamedTemporaryFile() as temp_output_file:
        for bundle_file in bundle.files:
            # Preprocess each file and copy onto temp_output_file
            bundle_file_output = processor_pipeline(bundle_file.processors, open(bundle_file.file_path, 'rb'))
            shutil.copyfileobj(bundle_file_output, temp_output_file)
            bundle_file_output.close()
            temp_output_file.write('\n')

        # Seek back to the start to run post processors
        temp_output_file.seek(0)

        # Post process the concatenated bundle
        processed_temp_output_file = processor_pipeline(bundle.processors, temp_output_file)
        processed_temp_output_file.seek(0)
        # Calculate a hash of the post processed file
        chunk_size = 2**14
        m = md5()
        while 1:
            chunk = processed_temp_output_file.read(chunk_size)
            if not chunk:
                break
            m.update(chunk)
        hash_version = m.hexdigest()

        processed_temp_output_file.seek(0)

        # Copy the file into its final location
        output_file_name = '%s.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), hash_version, bundle.bundle_type)
        with open(output_file_name, 'wb') as output_file:
            shutil.copyfileobj(processed_temp_output_file, output_file)

        processed_temp_output_file.close()

    return hash_version

class Command(BaseCommand):
    help = "Bundles up the media"

    def handle(self, *args, **options):
        self.stdout.write("Bundling...\n")

        _bundle_versions = {}
        set_bundle_versions(_bundle_versions)

        for bundle in get_bundles():
            self.stdout.write("Writing bundle: %s\n" % bundle.name)

            hash_version = make_bundle(bundle)

            # Build bundle versions as we're going along in case they're used in templated bundles
            _bundle_versions[bundle.name] = hash_version
            self.stdout.write("\t%s\n" % bundle.get_version())

        version_info = '\n'.join(['    "%s": "%s",' % version for version in _bundle_versions.iteritems()])

        with open(bundles_settings.BUNDLES_VERSION_FILE, 'wb') as bundles_versions:
            bundles_versions.write("""\
#!/usr/bin/env python

BUNDLES_VERSIONS = {
%s
}
""" % version_info)

        for single_file_input, single_file_output in bundles_settings.BUNDLES_SINGLE_FILES:
            self.stdout.write("Writing: %s\n" % single_file_output)
            file_type = os.path.splitext(single_file_input)[1][1:]
            processors = processor_library.get_default_preprocessors_for(file_type) + processor_library.get_default_postprocessors_for(file_type)
            with processor_pipeline(processors, open(single_file_input, 'rb')) as temp_output:
                with open(single_file_output, 'wb') as output_file:
                    shutil.copyfileobj(temp_output, output_file)

        self.stdout.write("Done.\n")



from django.core.management.base import BaseCommand, CommandError

from django_bundles.core import get_bundles
from django_bundles.processors import processor_pipeline
from django_bundles.utils import run_command
from django_bundles.conf.bundles_settings import bundles_settings

from optparse import make_option

import os


class Command(BaseCommand):
    help = "Lints the bundles based on settings.BUNDLES_LINTING"
    option_list = BaseCommand.option_list + (
        make_option('--failures-only',
            action='store_true',
            default=False,
            help='Only report failures'
        ),
    )

    def lint_file(self, filename, bundle_type, bundle_path):
        stdout, stderr = run_command(bundles_settings.BUNDLES_LINTING[bundle_type]['command'] % {
            'infile': filename
        })

        if (bundles_settings.BUNDLES_LINT_SUCCESS_OK and stdout.strip() == 'OK') or (not bundles_settings.BUNDLES_LINT_SUCCESS_OK and not stdout.strip()):
            if self.show_successes:
                self.stdout.write('OK\t\t%s\n' % bundle_path)
            failures = 0
        else:
            failures = 1
            self.stdout.write('FAIL\t\t%s\n' % bundle_path)
            self.stdout.write(stdout)

        return failures

    def handle(self, *args, **options):
        self.show_successes = not bool(options.get('failures_only'))

        failures = 0
        for bundle in get_bundles():
            for bundle_file in bundle.files:
                if bundle_file.lint:
                    with processor_pipeline(bundle_file.processors, open(bundle_file.file_path, 'rb'), require_actual_file=True) as file_output:
                        file_output.seek(0)
                        failures += self.lint_file(file_output.name, bundle.bundle_type, bundle_file.file_path)

        for single_file_path, _ in bundles_settings.BUNDLES_SINGLE_FILES:
            failures += self.lint_file(single_file_path, os.path.splitext(single_file_path)[1][1:], single_file_path)

        if failures:
            raise CommandError('%s FILE%s FAILED' % (failures, 'S' if failures > 1 else ''))
        else:
            self.stdout.write('\nALL FILES PASSED\n')



from django.views.debug import get_safe_settings
from django.template import Context, Template

from django_bundles.core import get_bundles
from django_bundles.processors import StringProcessor

class DjangoTemplateProcessor(StringProcessor):
    """
    Processor that runs Django's templating code across the file - 'settings' and 'bundles' are passed into context
    Override this class to add other things to context
    """
    def get_context(self):
        """
        Override this to change context
        """
        return {
            'settings': get_safe_settings(),
            'bundles': get_bundles(),
        }

    def process_string(self, input):
        template = Template(input)
        return template.render(Context(self.get_context()))




from django.conf import settings

USE_BUNDLES = not settings.DEBUG

DEFAULT_PREPROCESSORS = {
    'less': [
        'django_bundles.contrib.processors.less.LessProcessor',
    ],
}

DEFAULT_POSTPROCESSORS = {
    'js': [
        'django_bundles.contrib.processors.uglify.UglifyProcessor',
    ],
}

BUNDLES_LINTING = {}

BUNDLES_LINT_SUCCESS_OK = True

BUNDLES_SINGLE_FILES = ()



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from django_bundles.conf import default_settings

class SettingsHelper(object):
    USE_BUNDLES = getattr(settings, 'USE_BUNDLES', default_settings.USE_BUNDLES)

    DEFAULT_PREPROCESSORS = getattr(settings, 'DEFAULT_PREPROCESSORS', default_settings.DEFAULT_PREPROCESSORS)

    DEFAULT_POSTPROCESSORS = getattr(settings, 'DEFAULT_POSTPROCESSORS', default_settings.DEFAULT_POSTPROCESSORS)

    BUNDLES_LINTING = getattr(settings, 'BUNDLES_LINTING', default_settings.BUNDLES_LINTING)

    BUNDLES_LINT_SUCCESS_OK = getattr(settings, 'BUNDLES_LINT_SUCCESS_OK', default_settings.BUNDLES_LINT_SUCCESS_OK)

    BUNDLES_SINGLE_FILES = getattr(settings, 'BUNDLES_SINGLE_FILES', default_settings.BUNDLES_SINGLE_FILES)

    def __getattr__(self, name):
        if hasattr(settings, name):
            return getattr(settings, name)
        raise ImproperlyConfigured, "%s is a required setting for django_bundles" % name

bundles_settings = SettingsHelper()



from django_bundles.processors import ExecutableProcessor
from django_bundles.conf.bundles_settings import bundles_settings

class LessProcessor(ExecutableProcessor):
    command = bundles_settings.BUNDLES_LESS_COMMAND



from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.utils import run_command, get_class

from StringIO import StringIO
from tempfile import NamedTemporaryFile

class Processor(object):
    """
    Definition of a file processor
    """
    requires_actual_file = False # Whether this processor needs to be passed an actual file on disk or not (StringIO might be used otherwise)

    def process_file(self, input_file):
        """
        Processes the file and returns a new file(-like) object - override this to actually do something
        """
        return input_file

class ExecutableProcessor(Processor):
    """
    A Processor that runs an external command on the file
    """
    requires_actual_file = True
    command = 'cat %(infile)s > %(outfile)s' # Define a command

    def process_file(self, input_file):
        output_file = NamedTemporaryFile()
        run_command(self.command % {'infile': input_file.name, 'outfile': output_file.name})
        return output_file

class StringProcessor(Processor):
    """
    Processor that acts using strings in some way - abstracts opening files etc
    """
    def process_string(self, input):
        """
        This should be overridden to do string processing
        """
        return input

    def process_file(self, input_file):
        output_file = StringIO()
        output_file.write(self.process_string(input_file.read()))
        return output_file

def make_actual_file(input_file):
    if not hasattr(input_file, 'name'):
        new_input_file = NamedTemporaryFile()
        new_input_file.write(input_file.read())
        input_file.close()
        input_file = new_input_file
        input_file.seek(0)
    return input_file

def processor_pipeline(processors, start_file, require_actual_file=False):
    """
    Runs a list of processors over a file and returns a file(-like) object for the processed output
    """
    input_file = start_file

    for processor in processors:
        if processor:
            input_file.seek(0)

            if processor.requires_actual_file:
                input_file = make_actual_file(input_file)

            output_file = processor.process_file(input_file)
            input_file.close()
            input_file = output_file

    if require_actual_file:
        input_file = make_actual_file(input_file)

    return input_file

class ProcessorLibrary(object):
    def get_processor(self, processor_name):
        """
        Gets an instance of a processor from a string, class or instance
        """
        if isinstance(processor_name, Processor):
            return processor_name
        if isinstance(processor_name, type) and issubclass(processor_name, Processor):
            return processor_name()
        processor_class = get_class(processor_name)
        if processor_class:
            return processor_class()
        return None

    def get_processors(self, processors):
        return [instance for instance in (self.get_processor(processor) for processor in processors) if instance]

    def get_default_preprocessors_for(self, file_type):
        if file_type in bundles_settings.DEFAULT_PREPROCESSORS:
            return self.get_processors(bundles_settings.DEFAULT_PREPROCESSORS[file_type])
        return []

    def get_default_postprocessors_for(self, bundle_type):
        if bundle_type in bundles_settings.DEFAULT_POSTPROCESSORS:
            return self.get_processors(bundles_settings.DEFAULT_POSTPROCESSORS[bundle_type])
        return []

processor_library = ProcessorLibrary()


from django import template
from django.template.loader import render_to_string

from django.core.exceptions import ImproperlyConfigured

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.core import get_bundles

register = template.Library()

@register.simple_tag
def render_bundle(bundle_name):
    """
    Renders the HTML for a bundle in place - one HTML tag or many depending on settings.USE_BUNDLES
    """
    try:
        bundle = get_bundles()[bundle_name]
    except KeyError:
        raise ImproperlyConfigured("Bundle '%s' is not defined" % bundle_name)

    if bundles_settings.USE_BUNDLES:
        # Render one tag
        context = {
            'file_url': bundle.get_url(),
        }
        if bundle.media:
            context['attrs'] = { 'media': bundle.media }
        return render_to_string('django_bundles/%s.html' % bundle.bundle_type, context)

    # Render files individually
    bundle_files = []

    for bundle_file in bundle.files:
        context = {
            'file_url': bundle_file.file_url
        }
        if bundle_file.media:
            context['attrs'] = { 'media': bundle_file.media, }
        bundle_files.append(render_to_string('django_bundles/%s.html' % bundle_file.file_type, context))

    return '\n'.join(bundle_files)

if hasattr(register, 'assignment_tag'):
    @register.assignment_tag(name='get_bundles')
    def do_get_bundles():
        """
        Assigns the bundle definitions to a context variable
        """
        return get_bundles()



import shlex
import subprocess
import os
import fnmatch

def run_command(cmd):
    """
    Runs a command - including commands with pipes - and returns (stdout, stderr)
    """
    splitcmd = cmd.split('|')
    cmds = [subprocess.Popen(shlex.split(splitcmd[0]), stdout=subprocess.PIPE, stderr=subprocess.PIPE)]
    for i, c in enumerate(splitcmd[1:]):
        cmds.append(subprocess.Popen(shlex.split(c), stdin=cmds[i - 1].stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE))
    return cmds[-1].communicate()


def get_class(class_string):
    """
    Get a class from a dotted string
    """
    split_string = class_string.encode('ascii').split('.')
    import_path = '.'.join(split_string[:-1])
    class_name = split_string[-1]

    if class_name:
        try:
            if import_path:
                mod = __import__(import_path, globals(), {}, [class_name])
                cls = getattr(mod, class_name)
            else:
                cls = __import__(class_name, globals(), {})
            if cls:
                return cls
        except (ImportError, AttributeError):
            pass

    return None

def expand_file_names(path, files_root):
    """
    Expands paths (e.g. css/*.css in files_root /actual/path/to/css/files/)
    """
    dir_path, filename = os.path.split(path)
    return [os.path.join(dir_path, f) for f in fnmatch.filter(os.listdir(os.path.join(files_root, dir_path)), filename)]
